
// sunlight_map.cc -- part of gkrell_sunlight
//
// This program is copyright (C) 2008 Mauro Persano, and is free
// software which is freely distributable under the terms of the
// GNU public license, included as the file COPYING in this
// distribution.  It is NOT public domain software, and any
// redistribution not permitted by the GNU General Public License is
// expressly forbidden without prior written permission from
// the author.

#include <cmath>

#include "image.h"
#include "sunlight_map.h"

#ifdef USE_PNG
static const char *EARTH_TEXTURE = "earth.png";
#else
#include "earth.xpm"
#endif

sunlight_map::sunlight_map(int width)
#ifdef USE_PNG
  : earth_texture(image::from_png(EARTH_TEXTURE))
#else
  : earth_texture(image::from_xpm_data(earth))
#endif
  , cur_width(width)
{
	initialize_sunlight_image();
}

sunlight_map::~sunlight_map()
{
	delete earth_texture;
	delete sunlight_image;
}

void
sunlight_map::initialize_sunlight_image()
{
	cur_height = cur_width*earth_texture->height/
		earth_texture->width;

	sunlight_image = new image(cur_width, cur_height);

	float a = -.5f*M_PI;
	float da = M_PI/cur_height;

	phi_sin.clear();
	phi_cos.clear();

	for (int i = 0; i < cur_height; i++) {
		phi_sin.push_back(sin(a));
		phi_cos.push_back(cos(a));
		a += da;
	}

	float b = -M_PI;
	float db = (2*M_PI)/cur_width;

	theta_sin.clear();
	theta_cos.clear();

	for (int i = 0; i < cur_width; i++) {
		theta_sin.push_back(sin(b));
		theta_cos.push_back(cos(b));
		b += db;
	}
}

void
sunlight_map::set_width(int width)
{
	cur_width = width;
	delete sunlight_image;
	initialize_sunlight_image();
}

vector3
sunlight_map::sun_vector(int julian_date, int gmt_seconds) const
{
	// declination: angle between earth-sun line and equatorial plane;
	// 23.45 degrees (.4091 radians) on solstices, 0 on equinoxes
	float declination = .4091*sin((julian_date + 284.)*2*M_PI/365);

	// hour_angle: angular distance that earth has rotated in a day */
	float hour_angle = -static_cast<float>(gmt_seconds)*2.*M_PI/(24*60*60);

	return vector3(
		cos(declination)*cos(hour_angle),
		cos(declination)*sin(hour_angle),
		sin(declination));
}

void
sunlight_map::update(time_t t)
{
	const float MIN_INTENSITY = .4f;
	const float TRANSITION_THRESHOLD = .1f;

	struct tm *now = gmtime(&t);

	int julian_date = now->tm_yday;
	int gmt_seconds = now->tm_hour*60*60 + now->tm_min*60 + now->tm_sec;

	vector3 s = sun_vector(julian_date, gmt_seconds);

	rgb *dest = sunlight_image->data;

	const int texture_width = earth_texture->width;
	const int texture_height = earth_texture->height;

	unsigned v = 0;
	unsigned dv = ((texture_height - 1) << 16)/cur_height;

	for (int i = 0; i < cur_height; i++) {
		rgb *source = &earth_texture->data[(v >> 16)*texture_width];

		unsigned u = 0;
		unsigned du = ((texture_width - 1) << 16)/cur_width;

		for (int j = 0; j < cur_width; j++) {
			vector3 p(
				phi_cos[i]*theta_cos[j],
				phi_cos[i]*theta_sin[j],
				phi_sin[i]);

			float d = -p.dot_product(s);

			if (d < 0)
				d = MIN_INTENSITY;
			else if (d < TRANSITION_THRESHOLD)
				d = MIN_INTENSITY + (1.f - MIN_INTENSITY)*
				  (d/TRANSITION_THRESHOLD);
			else
				d = 1;

			const rgb *o = &source[u >> 16];

#ifdef USE_BILINEAR_FILTERING
			const unsigned uw1 = u & 0xffff;
			const unsigned uw0 = 0x10000 - uw1;

			const int r0 = (o[0].r*uw0 + o[1].r*uw1) >> 16;
			const int g0 = (o[0].g*uw0 + o[1].g*uw1) >> 16;
			const int b0 = (o[0].b*uw0 + o[1].b*uw1) >> 16;

			const int r1 = (o[texture_width].r*uw0 +
			  o[1 + texture_width].r*uw1) >> 16;
			const int g1 = (o[texture_width].g*uw0 +
			  o[1 + texture_width].g*uw1) >> 16;
			const int b1 = (o[texture_width].b*uw0 +
			  o[1 + texture_width].b*uw1) >> 16;

			const unsigned vw1 = v & 0xffff;
			const unsigned vw0 = 0x10000 - vw1;

			const int r = (r0*vw0 + r1*vw1) >> 16;
			const int g = (g0*vw0 + g1*vw1) >> 16;
			const int b = (b0*vw0 + b1*vw1) >> 16;
#else
			const int r = o->r;
			const int g = o->g;
			const int b = o->b;
#endif
			dest->r = static_cast<unsigned char>(r*d);
			dest->g = static_cast<unsigned char>(g*d);
			dest->b = static_cast<unsigned char>(b*d);

			++dest;
			u += du;
		}

		v += dv;
	}
}

void
sunlight_map::to_ppm(FILE *out)
{
	sunlight_image->to_ppm(out);
}
